const DIRECTIVE_PREFIX = 'l-';
const COMPONENT_FLAG = 'component';
const FOR_TEMPLATE_FLAG = '__for_template';
const MODEL_REGISTERED_FLAG = '__model_registered';
var DIRECTIVE_SHORTHANDS;
(function (DIRECTIVE_SHORTHANDS) {
    DIRECTIVE_SHORTHANDS["@"] = "on";
    DIRECTIVE_SHORTHANDS[":"] = "bind";
})(DIRECTIVE_SHORTHANDS || (DIRECTIVE_SHORTHANDS = {}));

var ASTNodeType;
(function (ASTNodeType) {
    ASTNodeType[ASTNodeType["NULL"] = -1] = "NULL";
    ASTNodeType[ASTNodeType["STATIC"] = 0] = "STATIC";
    ASTNodeType[ASTNodeType["DYNAMIC"] = 1] = "DYNAMIC";
})(ASTNodeType || (ASTNodeType = {}));

const rawDirectiveSplitRE = () => /:|\./gim;
const eventDirectivePrefixRE = () => /on|@/gim;
const parenthesisWrapReplaceRE = () => /\(|\)/gim;
const hasDirectiveRE = () => {
    return new RegExp(`(${DIRECTIVE_PREFIX}|${Object.keys(DIRECTIVE_SHORTHANDS).join('|')})\\w+`, 'gim');
};
const expressionPropRE = (prop) => {
    return new RegExp(`\\b${prop}\\b`, 'gim');
};

const resolveStateInExpression = (unresolvedExpression, deps = []) => {
    let expression = unresolvedExpression;
    deps.forEach((dep) => {
        if (dep !== expression)
            expression = expression.replace(expressionPropRE(dep), `$state.${dep}`);
    });
    return expression;
};
const computeExpression = (expression, el, returnable = true, refs = {}, deps) => {
    const formattedExpression = `${returnable ? `return ${expression}` : expression}`;
    const resolvedExpression = resolveStateInExpression(formattedExpression, deps);
    const specialPropertiesNames = ['$state', '$el', '$emit', '$event', '$refs'];
    const computeFunction = new Function(`'use strict';return function(${specialPropertiesNames.join(',')}){${resolvedExpression}}`)();
    const emit = (name, options, dispatchGlobal = true) => {
        const event = new CustomEvent(name, options);
        const target = dispatchGlobal ? window : el || window;
        target.dispatchEvent(event);
    };
    return (state, event) => {
        try {
            const value = state[expression];
            if (value) {
                return typeof value === 'function' ? value.bind(state)() : value;
            }
            else {
                return computeFunction(state, el, emit, event, refs);
            }
        }
        catch (err) {
            console.warn(`Lucia Error: "${err}"\n\nExpression: "${expression}"\nElement:`, el);
        }
    };
};

const getElementCustomProp = (el, prop) => {
    return el[prop];
};
const setElementCustomProp = (el, prop, value) => {
    el[prop] = value;
};

const removeDupesFromArray = (array) => {
    return [...new Set(array)];
};

const isListRenderScope = (el) => {
    return el.hasAttribute(`${DIRECTIVE_PREFIX}for`);
};
const isUnderListRenderScope = (el) => {
    return !!el.parentElement && el.parentElement.hasAttribute(`${DIRECTIVE_PREFIX}for`);
};
const createASTNode = (el, state) => {
    const [directives, deps] = collectAndInitDirectives(el, state);
    const hasDirectives = Object.keys(directives).length > 0;
    const hasDepInDirectives = Object.values(directives).some(({ value }) => Object.keys(state).some((prop) => expressionPropRE(prop).test(value)));
    const type = hasDepInDirectives ? ASTNodeType.DYNAMIC : ASTNodeType.STATIC;
    const node = { el, deps, directives, type };
    return hasDirectives ? node : undefined;
};
const collectRefs = () => {
    const refDirective = `${DIRECTIVE_PREFIX}ref`;
    const refElements = document.querySelectorAll(`[${refDirective}]`);
    const refs = {};
    refElements.forEach((refElement) => {
        const name = refElement.getAttribute(refDirective);
        if (name)
            refs[name] = refElement;
    });
    return refs;
};
const collectAndInitDirectives = (el, state = {}) => {
    const directives = {};
    const refs = collectRefs();
    const nodeDeps = [];
    for (const { name, value } of el.attributes) {
        const isStateDirective = name === `${DIRECTIVE_PREFIX}state`;
        const hasDirectivePrefix = name.startsWith(DIRECTIVE_PREFIX);
        const hasDirectiveShorthandPrefix = Object.keys(DIRECTIVE_SHORTHANDS).some((shorthand) => name.startsWith(shorthand));
        if (isStateDirective || !(hasDirectivePrefix || hasDirectiveShorthandPrefix))
            continue;
        const depsInFunctions = [];
        const propsInState = Object.keys(state);
        let returnable = true;
        const deps = propsInState.filter((prop) => {
            const hasDep = expressionPropRE(prop).test(String(value));
            if (hasDep && typeof state[prop] === 'function') {
                const depsInFunction = propsInState.filter((p) => {
                    return expressionPropRE(p).test(String(state[prop]));
                });
                depsInFunctions.push(...depsInFunction);
            }
            return hasDep;
        });
        if (eventDirectivePrefixRE().test(name))
            returnable = false;
        if (name.includes('for') && getElementCustomProp(el, FOR_TEMPLATE_FLAG) === undefined) {
            setElementCustomProp(el, FOR_TEMPLATE_FLAG, String(el.innerHTML).trim());
            returnable = false;
        }
        const uniqueCompiledDeps = removeDupesFromArray([...deps, ...depsInFunctions]);
        nodeDeps.push(...uniqueCompiledDeps);
        const directiveData = {
            compute: computeExpression(value, el, returnable, refs, uniqueCompiledDeps),
            deps: uniqueCompiledDeps,
            value,
        };
        const directiveName = hasDirectivePrefix
            ? name.slice(DIRECTIVE_PREFIX.length)
            : `${DIRECTIVE_SHORTHANDS[name[0]]}:${name.slice(1)}`;
        directives[directiveName.toLowerCase()] = directiveData;
    }
    return [directives, removeDupesFromArray(nodeDeps)];
};
const flattenElementChildren = (rootElement, isListGroup = false, ignoreRootElement = false) => {
    const collection = [];
    const isList = isListRenderScope(rootElement);
    const isUnderList = isUnderListRenderScope(rootElement);
    if (!isListGroup && (isList || isUnderList))
        return collection;
    if (!ignoreRootElement && (!isListGroup || !isList))
        collection.push(rootElement);
    if (isListGroup || (!isList && !isUnderList)) {
        for (const childElement of rootElement.children) {
            if (childElement instanceof HTMLElement) {
                if (!isListGroup && isListRenderScope(childElement)) {
                    collection.push(childElement);
                }
                else {
                    if (childElement.hasAttribute(`${DIRECTIVE_PREFIX}state`))
                        continue;
                    collection.push(...flattenElementChildren(childElement, isListGroup, childElement.attributes.length === 0));
                }
            }
        }
    }
    return collection;
};
const compile = (el, state = {}, ignoreRootElement = false) => {
    const ast = [];
    const isListGroup = getElementCustomProp(el, COMPONENT_FLAG) !== undefined && isListRenderScope(el);
    const elements = flattenElementChildren(el, isListGroup, ignoreRootElement);
    const maskDirective = `${DIRECTIVE_PREFIX}mask`;
    elements.forEach((element) => {
        if (element.hasAttribute(maskDirective)) {
            element.removeAttribute(maskDirective);
        }
        if (hasDirectiveRE().test(element.outerHTML)) {
            const newASTNode = createASTNode(element, state);
            if (newASTNode)
                ast.push(newASTNode);
        }
    });
    return ast;
};

const formatAcceptableWhitespace = (expression) => {
    const whitespaceRE = /\s+/gim;
    return expression.replace(whitespaceRE, ' ').trim();
};
const bindDirective = ({ el, parts, data, state }) => {
    switch (parts[1]) {
        case 'class': {
            const classes = data.compute(state);
            if (typeof classes === 'string') {
                return el.setAttribute('class', formatAcceptableWhitespace(`${el.className} ${classes}`));
            }
            else if (Array.isArray(classes)) {
                return el.setAttribute('class', formatAcceptableWhitespace(`${el.className} ${classes.join(' ')}`));
            }
            else {
                const activeClasses = [];
                Object.entries(classes).forEach(([className, classValue]) => {
                    if (classValue)
                        activeClasses.push(className);
                });
                const removeDynamicClassesRE = new RegExp(`\\b${Object.keys(classes).join('|')}\\b`, 'gim');
                const rawClasses = el.className.replace(removeDynamicClassesRE, '');
                if (activeClasses.length > 0) {
                    return el.setAttribute('class', formatAcceptableWhitespace(`${rawClasses} ${activeClasses.join(' ')}`));
                }
                else if (formatAcceptableWhitespace(rawClasses).length > 0) {
                    return el.setAttribute('class', formatAcceptableWhitespace(rawClasses));
                }
                else if (el.hasAttribute('class')) {
                    return el.removeAttribute('class');
                }
            }
            break;
        }
        case 'style': {
            const styles = data.compute(state);
            if (el.hasAttribute('style'))
                el.removeAttribute('style');
            Object.entries(styles).forEach(([styleName, styleValue]) => {
                el.style[styleName] = styleValue;
            });
            break;
        }
        default: {
            const attributes = data.compute(state);
            if (typeof attributes === 'object' && attributes !== null) {
                Object.entries(attributes).forEach(([name, value]) => {
                    if (value) {
                        el.setAttribute(name, value);
                    }
                    else {
                        el.removeAttribute(name);
                    }
                });
            }
            else if (attributes) {
                el.setAttribute(parts[1], attributes);
            }
            else {
                if (el.hasAttribute(parts[1]))
                    el.removeAttribute(parts[1]);
            }
            break;
        }
    }
};

const lazy = (threshold, generatorFunction) => {
    const generator = generatorFunction();
    return function next() {
        const start = performance.now();
        let task = null;
        do {
            task = generator.next();
        } while (performance.now() - start < threshold && !task.done);
        if (task.done)
            return;
        setTimeout(next);
    };
};

const render = (ast, directives, state = {}, changedProps = []) => {
    const legalDirectiveNames = Object.keys(directives);
    const LAZY_MODE_TIMEOUT = 25;
    lazy(LAZY_MODE_TIMEOUT, function* () {
        for (const node of ast) {
            if (node.type === ASTNodeType.NULL)
                continue;
            const isStatic = node.type === ASTNodeType.STATIC;
            if (isStatic)
                node.type = ASTNodeType.NULL;
            yield;
            const nodeHasDep = changedProps.some((prop) => node.deps.includes(prop));
            if (!nodeHasDep && !isStatic)
                continue;
            for (const [directiveName, directiveData] of Object.entries(node.directives)) {
                const rawDirectiveName = directiveName.split(rawDirectiveSplitRE())[0];
                if (!legalDirectiveNames.includes(rawDirectiveName.toUpperCase()))
                    continue;
                yield;
                const directiveHasDep = changedProps.some((prop) => directiveData.deps.includes(prop));
                const isMaskDirective = directiveName === `${DIRECTIVE_PREFIX}mask`;
                const isStaticDirective = Object.keys(directiveData.deps).length === 0;
                if (directiveHasDep || isStatic || isStaticDirective) {
                    const directiveProps = {
                        el: node.el,
                        parts: directiveName.split(rawDirectiveSplitRE()),
                        data: directiveData,
                        node,
                        state,
                    };
                    renderDirective(directiveProps, directives);
                    if (isStaticDirective || isMaskDirective) {
                        delete node.directives[directiveName];
                        if (isMaskDirective) {
                            node.el.removeAttribute(`${DIRECTIVE_PREFIX}mask`);
                        }
                    }
                }
            }
            if (node.directives['on:effect']) {
                const effectEvent = new CustomEvent('effect');
                node.el.dispatchEvent(effectEvent);
            }
        }
    })();
};

const adjustDeps = (ast, currentDeps, node, directiveName) => {
    const deps = [];
    ast.forEach((node) => {
        deps.push(...node.deps);
    });
    const cleanedDeps = removeDupesFromArray([...currentDeps, ...deps]);
    node.deps = cleanedDeps;
    node.directives[directiveName].deps = cleanedDeps;
};

const forDirective = ({ el, data, state, node }) => {
    var _a;
    const originalAST = getElementCustomProp(el, COMPONENT_FLAG);
    if (!originalAST)
        setElementCustomProp(el, COMPONENT_FLAG, compile(el, state));
    const forLoopRE = /\s+(?:in|of)\s+/gim;
    const [expression, target] = data.value.split(forLoopRE);
    const [item, index] = expression === null || expression === void 0 ? void 0 : expression.trim().replace(parenthesisWrapReplaceRE(), '').split(',');
    const currArray = (_a = state[target === null || target === void 0 ? void 0 : target.trim()]) !== null && _a !== void 0 ? _a : computeExpression(target === null || target === void 0 ? void 0 : target.trim(), el, true)(state);
    const template = getElementCustomProp(el, FOR_TEMPLATE_FLAG);
    if (el.innerHTML.trim() === template)
        el.innerHTML = '';
    const arrayDiff = (currArray === null || currArray === void 0 ? void 0 : currArray.length) - el.children.length;
    if ((currArray === null || currArray === void 0 ? void 0 : currArray.length) === 0)
        el.innerHTML = '';
    else if (arrayDiff !== 0) {
        for (let i = Math.abs(arrayDiff); i > 0; --i) {
            if (arrayDiff < 0)
                el.removeChild(el.lastChild);
            else {
                let content = String(template);
                const isTable = /^[^\S]*?<(t(?:head|body|foot|r|d|h))/i.test(content);
                if (item) {
                    content = content.replace(expressionPropRE(`this\\.${item.trim()}`), `${target}[${currArray.length - i}]`);
                }
                if (index) {
                    content = content.replace(expressionPropRE(`this\\.${index.trim()}`), String(currArray.length - i));
                }
                if (isTable)
                    content = `<table>${content}</table>`;
                const fragment = document.createRange().createContextualFragment(content);
                el.appendChild(isTable ? fragment.firstElementChild.firstElementChild : fragment);
            }
        }
        setElementCustomProp(el, COMPONENT_FLAG, compile(el, state));
    }
    if (!originalAST) {
        adjustDeps(getElementCustomProp(el, COMPONENT_FLAG), data.deps, node, 'for');
        el.removeAttribute(`${DIRECTIVE_PREFIX}for`);
    }
    const ast = arrayDiff === 0 ? originalAST : compile(el, state, true);
    render(ast || [], directives, state, node.deps);
};

const htmlDirective = ({ el, data, state, node }) => {
    var _a;
    node = node;
    const marker = getElementCustomProp(el, COMPONENT_FLAG);
    const ret = (_a = data.compute(state)) !== null && _a !== void 0 ? _a : data.value;
    if (ret !== el.innerHTML) {
        el.innerHTML = ret;
        if (hasDirectiveRE().test(ret)) {
            const ast = marker !== null && marker !== void 0 ? marker : compile(el, state, true);
            if (!marker)
                adjustDeps(ast, data.deps, node, 'html');
            render(ast, directives, state, data.deps);
            setElementCustomProp(el, COMPONENT_FLAG, ast);
        }
    }
};

const inputCallback = (el, hydratedValue, data, state) => {
    if (el.type === 'checkbox') {
        el.value = String(el.checked);
    }
    const isNumber = typeof hydratedValue === 'number' && !isNaN(el.value);
    const isBoolean = typeof hydratedValue === 'boolean' && (el.value === 'true' || el.value === 'false');
    const isNullish = (hydratedValue === null || hydratedValue === undefined) &&
        (el.value === 'null' || el.value === 'undefined');
    let payload;
    if (isNumber) {
        payload = Number(el.value);
    }
    else if (isBoolean) {
        payload = el.value === 'true';
    }
    else if (isNullish) {
        if (el.value === 'null')
            payload = null;
        else
            payload = undefined;
    }
    else {
        payload = String(el.value);
    }
    if (state[data.value]) {
        state[data.value] = payload;
    }
    else {
        payload = typeof payload === 'string' ? `'${payload}'` : payload;
        computeExpression(`$state.${data.value} = ${payload}`, el, true)(state);
    }
    return payload;
};
const modelDirective = ({ el: awaitingTypecastEl, parts, data, state, }) => {
    var _a;
    const el = awaitingTypecastEl;
    const hydratedValue = (_a = state[data.value]) !== null && _a !== void 0 ? _a : computeExpression(data.value, el, true)(state);
    const accessor = el.type === 'checkbox' ? 'checked' : 'value';
    if (el[accessor] !== String(hydratedValue)) {
        el[accessor] = hydratedValue;
    }
    if (!getElementCustomProp(el, MODEL_REGISTERED_FLAG)) {
        const callback = () => inputCallback(el, hydratedValue, data, state);
        el.addEventListener(parts[1] === 'debounce' ? 'change' : 'input', callback);
        setElementCustomProp(el, MODEL_REGISTERED_FLAG, true);
    }
};

const onDirective = ({ el, parts, data, state }) => {
    const options = {};
    const globalScopeEventProps = ['outside', 'global'];
    const eventProps = parts.slice(2);
    if (getElementCustomProp(el, `__on_${parts[1]}_registered`))
        return;
    const target = globalScopeEventProps.some((prop) => String(eventProps).includes(prop))
        ? window
        : el;
    const handler = (event) => {
        if (event instanceof KeyboardEvent && /\d/gim.test(String(eventProps))) {
            const whitelistedKeycodes = [];
            eventProps.forEach((eventProp) => {
                if (!isNaN(eventProp)) {
                    whitelistedKeycodes.push(Number(eventProp));
                }
            });
            if (!whitelistedKeycodes.includes(event.keyCode))
                return;
        }
        if (eventProps.includes('prevent'))
            event.preventDefault();
        if (eventProps.includes('stop'))
            event.stopPropagation();
        if (eventProps.includes('self')) {
            if (event.target !== el)
                return;
        }
        if (eventProps.includes('outside')) {
            if (el.contains(event.target))
                return;
            if (el.offsetWidth < 1 && el.offsetHeight < 1)
                return;
        }
        data.compute(state, event);
    };
    options.once = eventProps.includes('once');
    options.passive = eventProps.includes('passive');
    target.addEventListener(parts[1], handler, options);
    setElementCustomProp(el, `__on_${parts[1]}_registered`, true);
};

const showDirective = ({ el, data, state }) => {
    const ret = data.compute(state);
    if (ret !== el.style.display)
        el.style.display = ret ? '' : 'none';
    if (el.style.length === 0)
        el.removeAttribute('style');
};

const textDirective = ({ el, data, state }) => {
    var _a;
    const ret = (_a = data.compute(state)) !== null && _a !== void 0 ? _a : data.value;
    if (ret !== el.textContent) {
        el.textContent = ret;
    }
};

const directives = {
    BIND: bindDirective,
    HTML: htmlDirective,
    MODEL: modelDirective,
    SHOW: showDirective,
    ON: onDirective,
    TEXT: textDirective,
    FOR: forDirective,
};
const renderDirective = (props, directives) => {
    directives[props.parts[0].toUpperCase()](props);
};

const arrayEquals = (firstArray, secondArray) => {
    return (Array.isArray(firstArray) &&
        Array.isArray(secondArray) &&
        firstArray.length === secondArray.length &&
        firstArray.every((value, i) => value === secondArray[i]));
};
const reactive = (state, render, watchers = {}) => {
    const handler = {
        get(target, key) {
            const ret = target[key];
            if (typeof ret === 'object' && ret !== null) {
                return new Proxy(ret, handler);
            }
            else {
                return ret;
            }
        },
        set(target, key, value) {
            const hasArrayMutationKey = !isNaN(Number(key)) || key === 'length';
            const props = hasArrayMutationKey ? [] : [key];
            if (Array.isArray(target) && hasArrayMutationKey) {
                const keys = Object.keys(state).filter((prop) => {
                    return (Array.isArray(state[prop]) && arrayEquals(state[prop], target));
                });
                props.push(...keys);
            }
            else {
                if (!Object.is(target, state)) {
                    const keys = Object.keys(state).filter((prop) => {
                        return (Object.prototype.toString.call(state[prop]) === '[object Object]' &&
                            JSON.stringify(state[prop]).indexOf(key) > -1);
                    });
                    props.push(...keys);
                }
            }
            target[key] = value;
            render(props);
            Object.entries(watchers).forEach(([prop, watcher]) => {
                if (props.includes(prop))
                    watcher();
            });
            return true;
        },
    };
    return new Proxy(Object.seal(state), handler);
};

class Component {
    constructor(state = {}, directives = {}, watchers = {}) {
        this.ast = [];
        this.state = state;
        this.directives = directives;
        this.watchers = watchers;
    }
    mount(el) {
        const rootEl = typeof el === 'string'
            ? document.querySelector(el) || document.body
            : el;
        const newState = { ...this.state, $render: this.render.bind(this) };
        this.ast = compile(rootEl, this.state);
        this.directives = { ...this.directives, ...directives };
        this.state = reactive(newState, this.render.bind(this), this.watchers);
        this.render();
        setElementCustomProp(rootEl, COMPONENT_FLAG, this);
        const mountedEvent = new CustomEvent('mounted');
        rootEl.dispatchEvent(mountedEvent);
        return this.state;
    }
    directive(name, callback) {
        this.directives[name.toUpperCase()] = callback;
    }
    watch(name, callback) {
        this.watchers[name] = callback;
    }
    render(props = Object.keys(this.state)) {
        render(this.ast, directives, this.state, props);
    }
}
const component = (state) => new Component(state);

const init = (element = document) => {
    const stateDirective = `${DIRECTIVE_PREFIX}state`;
    const componentElements = element.querySelectorAll(`[${stateDirective}]`);
    const uninitializedComponents = [...componentElements].filter((el) => getElementCustomProp(el, COMPONENT_FLAG) === undefined);
    uninitializedComponents.forEach((uninitializedComponent) => {
        const stateExpression = uninitializedComponent.getAttribute(stateDirective);
        const state = new Function(`return ${stateExpression}`)() || {};
        const currentComponent = component(state);
        currentComponent.mount(uninitializedComponent);
    });
};

export { compile, component, computeExpression, directives, init, reactive, render };
